import lib.reg as reg
from lib.parsing.parser import Parser, ParsingError
from lib.checker import transform, dfa_transform, nfa_transform, check_task, check_alphabets
from typing import List, Tuple
from enum import Enum
import enum


class Size(Enum):
    Empty = enum.auto()
    Finite = enum.auto()
    Infinite = enum.auto()
    Universal = enum.auto()


class Language:  # contains information about formalism and its language
    def __init__(self, string: str, task: str = "DFA"):
        self.string = string
        self.task = task
        self.dfa = None
        self.minimal = None
        self.size = None
        self.sizename = None
        self.example = None
        self.gen_lang_html()

    def gen_lang_html(self):
        parser = Parser()
        dfa = dfa_transform(self.string, self.task)
        self.minimal = parser.dfa_to_str(dfa.minimize().canonize())

        empty = dfa.is_empty()
        if empty:
            size = Size.Empty
        else:
            self.example = empty.counterexample
            if not empty.inf:
                size = Size.Finite
            else:
                size = Size.Infinite
                if dfa.is_universal():
                    size = Size.Universal

        name = {"Empty": "prázdný",
                "Finite": "konečný",
                "Infinite": "nekonečný",
                "Universal": "univerzální"}
        self.size = size
        self.sizename = name[size.name]
        self.dfa = dfa

    def gen_is_task(self, task: str) -> str:
        teacher = "NFA" if self.task == "EFA" else self.task
        parser = Parser()
        dfa = parser.dfa_to_str(self.dfa.canonize())
        return f"{teacher}-{task}:{dfa}"


class WebChecker:
    def __init__(self, student_string: str, task: str):
        self.student_string = student_string
        self.task = task
        self.task_solved = ""
        self.alphabets = ""
        self.eq = None
        self.teacher = None
        self.student = None
        self.is_task = None
        self.img_src = None
        self.languages = [None, None, None, None]

    # nearly equivalent to dfa_task in fja_checker (for IS ROPOTs)
    def dfa_task(self, teacher_type, teacher_string, task, student_string):
        try:
            teacher = Language(string=teacher_string, task=teacher_type)
            student = Language(string=student_string, task=self.task)
            student_solution = transform(student_string, task)  # only for task checking

        except ParsingError as ex:  # Parsing error, don't continue
            return ex.args

        self.teacher = teacher
        self.student = student
        self.is_task = teacher.gen_is_task(task=self.task)

        if isinstance(student_solution, reg.DFA) or isinstance(student_solution, reg.NFA):
            self.task_solved = check_task(student_solution, task)

        if teacher.size == Size.Empty and student.size == Size.Empty:
            return self.task_solved == ""

        # if the alphabets aren't same, languages aren't equal (unless they are both empty)
        alphabets = check_alphabets(student_alpha=student.dfa.characters,
                                    teacher_alpha=teacher.dfa.characters, task=task)
        if alphabets != "":
            self.alphabets = alphabets

            return False

        self.eq = reg.DFA.is_equivalent(student.dfa, teacher.dfa)

        self.img_src = self.relation(self.eq.left_counterexample is None and
                                     self.eq.right_counterexample is None)
        if self.task_solved == "" and \
                self.eq.right_counterexample is None and \
                self.eq.left_counterexample is None:
            return True
        return False


    def compare(self, teacher_string: str, teacher_type: str):
        #signal.alarm(50)  # from fja_checker: here would kill the web
        try:
            return self.dfa_task(teacher_type=teacher_type, teacher_string=teacher_string,
                                 task=self.task, student_string=self.student_string)

        except ParsingError as ex:
            return ex.args
            #raise ParsingError(ex.args)


    def convert(self, student_type):
        try:
            parser = Parser()
            if self.task in {"DFA", "TOT", "MIN"}:
                dfa = dfa_transform(self.student_string, student_type)

                if self.task == "DFA":
                    return parser.dfa_to_str(dfa.canonize())
                if self.task == "TOT":
                    return parser.dfa_to_str(dfa.total().canonize())
                if self.task == "MIN":
                    return parser.dfa_to_str(dfa.minimize().canonize())

            nfa = nfa_transform(self.student_string, student_type)
            if self.task == "EFA":
                return parser.nfa_to_str(nfa)

            nfa = nfa.eliminate_epsilon()
            if self.task == "NFA":
                return parser.nfa_to_str(nfa)

            if self.task == "GRA":
                return parser.reggrammar_to_str(nfa.nfa_to_reggrammar().eliminate_useless())

        except ParsingError as ex:
            return ex.args
            # raise ParsingError(ex.args)

        except Exception as ex:
            print("Error inside of web checker:", ex.args)

    def relation(self, eq: bool) -> str:
        student = self.student.dfa
        teacher = self.teacher.dfa

        # language 0 on picture: complement of both
        self.languages[0] = self.language((reg.DFA.union(teacher, student)).complement())

        if not eq:
            if self.eq.left_counterexample is not None:  # extra word
                if self.eq.right_counterexample is not None:  # missing word
                    intersection = reg.DFA.intersection(student, teacher)
                    if intersection.is_empty():  # disjunction
                        self.languages[1] = self.teacher
                        self.languages[2] = self.student
                        return "disjunction"
                    else:
                        self.languages[1] = self.language(reg.DFA.subtraction(teacher, student))
                        self.languages[2] = self.language(reg.DFA.subtraction(student, teacher))
                        self.languages[3] = self.language(reg.DFA.intersection(teacher, student))
                        return "intersection"

                else:  # teacher is subset of student
                    self.languages[1] = self.teacher
                    self.languages[2] = self.language(reg.DFA.subtraction(student, teacher))
                    picture = "t_in_s"
            else:  # student is subset of teacher
                self.languages[1] = self.language(reg.DFA.subtraction(teacher, student))
                self.languages[2] = self.student
                picture = "s_in_t"

        else:
            #print(self.eq.left_counterexample, self.eq.right_counterexample)
            picture = "eq"
            self.languages[0] = self.language(teacher.complement())
            self.languages[1] = self.teacher

        #if self.teacher.size == Size.Empty or self.student.size == Size.Universal:
        #    picture += "_empty"
        if self.teacher.size == Size.Universal or self.student.size == Size.Universal:
            picture += "_universal"
            if self.teacher.size == Size.Universal:
                self.languages.pop(0)
            if self.student.size == Size.Universal:
                self.languages[0] = self.languages[2]
                self.languages[2] = None

        return picture  # name of static image file to show in feedback

    def language(self, dfa: reg.DFA) -> Language:
        parser = Parser()
        return Language(parser.dfa_to_str(dfa))

    def DFA_lines(self, str):
        pass


